#import "ImageView.h"
#import "../Utils/FileUtils.h"
#import "../Utils/Utils.h"
#import <QuartzCore/QuartzCore.h>


// TODO: implement zooming
#define MIN_ZOOM_FACTOR		0.1f
#define MAX_ZOOM_FACTOR		5.0f


@implementation ImageView

-(void)awakeFromNib
{
	DEBUG_LOG(@"ImageView - awakeFromNib");

	mImageURL = nil;
	
	// NOTE: When adding this class to your XIB, you can either drop an
	//		 IKImageView, or a simple NSView to your interface, and then change
	//		 the class type to ImageView. The "normal" way of doing this would
	//		 of course to drop an IKImageView, so that your can configure the
	//		 IKImageView through the interface builder. BUT ! If we do that, we
	//		 will have loooots of bugs, especially with overlay layers for
	//		 animated gifs. So, instead, I just use an NSView. And as a result,
	//		 I have to configure it programmatically here ...
	// NOTE: we need to setImageWithURL once ... It seems this method initialize
	//		 a lot of stuff in the IKImageView, and if we don't do that, the
	//		 following setBackgrondColor, etc. will have not effect until the
	//		 first setImageWithURL call occurs.

	[super setImageWithURL:nil];
	[self setBackgroundColor:[NSColor blackColor]];
	[self setAutoresizes:YES];
}

-(void)dealloc
{
	if (mImageURL != nil)
	{
		[mImageURL release];
		mImageURL = nil;
	}
	[super dealloc];
}

-(NSString *)currentImage
{
	return [mImageURL path];
}

-(void)setCurrentImage:(NSString *)image
{
	if (image != nil)
		[self setImageWithURL:[NSURL fileURLWithPath:image] force:NO];
	else
		[self setImageWithURL:nil force:NO];
}

-(void)setImageWithURL:(NSURL *)url force:(BOOL)force
{
	if (url == mImageURL && force == NO)
		return;

	DEBUG_LOG(@"ImageView - setImageWithURL:[%@] force:%@",
			 [url path],
			 force == YES ? @"YES" : @"NO");

	// re-set autoresize to YES (a previous zooming might have set
	// this to NO)
	[self setAutoresizes:YES];

	// checks if a layer is already set and remove all previous animations
	if ([self overlayForType:IKOverlayTypeImage] == nil)
		[self setOverlay:[CALayer layer] forType:IKOverlayTypeImage];
	[[self overlayForType:IKOverlayTypeImage] removeAllAnimations];

	// backup the image url
	if (url != nil)
	{
		if (mImageURL != url)
		{
			[mImageURL release];
			mImageURL = [url copy];
		}

		// set the image of the IKImageView. For a reason I don't know, if we
		// don't do that, the overlay layers will often crash ... Maybe this
		// method initialize some stuff in the IKImageView or whatever ...
		[super setImageWithURL:mImageURL];

		// check if the image a gif
		if ([FileUtils isGIF:[url path]] == YES)
		{
			// create the image
			NSImage * image = [[NSImage alloc] initWithContentsOfFile:[url path]];
			
			// get the image representations
			NSArray * reps = [image representations];
			for (NSImageRep * rep in reps)
			{
				// find the bitmap representation
				if ([rep isKindOfClass:[NSBitmapImageRep class]] == NO)
					continue;

				// get the bitmap representation
				NSBitmapImageRep * bitmapRep = (NSBitmapImageRep *)rep;
				
				// check multiframe gif
				int numFrame = [[bitmapRep valueForProperty:NSImageFrameCount] intValue];
				if (numFrame == 0)
					break;
				
				// create the layer, animation and value array
				CAKeyframeAnimation * animation = [CAKeyframeAnimation animationWithKeyPath:@"contents"];
				NSMutableArray * values = [NSMutableArray array];
				
				// loop through the frames
				float animationDuration = 0.0f;
				for (int i = 0; i < numFrame; ++i)
				{
					// set the current frame
					[bitmapRep setProperty:NSImageCurrentFrame withValue:[NSNumber numberWithInt:i]];
					
					// if the frame duration is 0, replace it by a default of 1/20 second
					if ([[bitmapRep valueForProperty:NSImageCurrentFrameDuration] floatValue] <= 0.000001f)
						[bitmapRep setProperty:NSImageCurrentFrameDuration withValue:[NSNumber numberWithFloat:1.0f / 20.0f]];
					
					// add the CGImageRef to this frame to the value array
					[values addObject:(id)[bitmapRep CGImage]];
					
					// update the duration of the animation
					animationDuration += [[bitmapRep valueForProperty:NSImageCurrentFrameDuration] floatValue];
				}
				
				// setup the animation
				[animation setValues:values];
				[animation setCalculationMode:@"discrete"];
				[animation setDuration:animationDuration];
				[animation setRepeatCount:HUGE_VAL];

				// create the layer and add the animation
				CALayer * layer = [CALayer layer];
				[layer addAnimation:animation forKey:@"contents"];
				[self setOverlay:layer forType:IKOverlayTypeImage];

				// we have our animated gif, no need to continue
				break;
			}

			// release the image
			[image release];
		}
	}
	else if (mImageURL != nil)
	{
		if (mImageURL != nil)
		{
			[mImageURL release];
			mImageURL = nil;
		}
		[super setImageWithURL:nil];
	}
}

-(void)toggleFullscreen
{
	if ([self isInFullScreenMode] == YES)
		[self leaveFullscreen];
	else
		[self enterFullscreen];
}

-(void)enterFullscreen
{
	if ([self isInFullScreenMode] == YES)
		return;

	DEBUG_LOG(@"ImageView - Enter fullscreen");

	// create a dictionary for the fullscreen options
	// NSFullScreenModeWindowLevel is used to display on top of everything
	// NSApplicationPresentationDefault is used so that we can still use Spaces, swtich process, etc.
	NSMutableDictionary * options = [NSMutableDictionary dictionary];
	[options setValue:[NSNumber numberWithInt:NSMainMenuWindowLevel + 1]
			 forKey:NSFullScreenModeWindowLevel];
	[options setValue:[NSNumber numberWithInt:NSApplicationPresentationDefault]
			 forKey:NSFullScreenModeApplicationPresentationOptions];
	
	// enter fullscreen mode and hide cursor
	[self enterFullScreenMode:[NSScreen mainScreen] withOptions:options];
	[NSCursor hide];
	
	// NOTE: when the image view is not in fullscreen, the image is quite small,
	//		 and when we go to fullscreen, instead of displaying the full
	//		 resolution image, the view only scales the current image, resulting
	//		 in low resolution view (if we change image, the problem doesn't
	//		 happen since the view loads the image at the correct resolution)
	//		 This is why we need to reload the image when we go to fullscreen,
	//		 in order to have a resolution adapted to the fullscreen resolution.
	[self reloadImage];
}

-(void)leaveFullscreen
{
	if ([self isInFullScreenMode] == NO)
		return;

	DEBUG_LOG(@"ImageView - Leave fullscreen");

	[self exitFullScreenModeWithOptions:nil];
	[NSCursor unhide];
}

-(void)reloadImage
{
	DEBUG_LOG(@"ImageView - reload image");

	[self setImageWithURL:mImageURL force:YES];
}

@end
